/*

Copyright (C) 2008 Michael Goffioul

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <map>

#include "classdef.h"
#include "defun-dld.h"
#include "ov-typeinfo.h"

static std::map<std::string, cdef_class> all_classes;
static std::map<std::string, cdef_object> all_packages;

static bool
check_access (const std::string& req, const std::string& acc)
{
  if (req == "private")
    return true;
  else if (req == "protected")
    return (acc != "private");
  else
    return (acc == "public");
}

static std::string
superclass_access (const std::string& acc)
{
  if (acc == "public")
    return acc;
  else
    return "protected";
}

static cdef_class
lookup_class (const std::string& name, bool error_if_not_found = true)
{
  std::map<std::string, cdef_class>::iterator it = all_classes.find (name);

  if (it == all_classes.end ())
    {
      // FIXME: should look into load-path
      if (error_if_not_found)
	error ("class not found: %s", name.c_str ());
    }
  else
    {
      cdef_class& cls = it->second;

      if (! cls.is_builtin ())
	{
	  // FIXME: check whether a class reload is needed
	}

      if (cls.ok ())
	return cls;
      else
	all_classes.erase (it);
    }

  return cdef_class ();
}

static bool
is_superclass (const cdef_class& clsa, const cdef_class& clsb,
	       bool allow_equal = true)
{
  if (allow_equal && clsa == clsb)
    return true;
  else
    {
      Cell c = clsb.get ("SuperClasses").cell_value ();

      bool retval = false;

      for (int i = 0; ! retval && i < c.numel (); i++)
	{
	  cdef_class cls = lookup_class (c(i).string_value ());

	  if (! error_state)
	    retval = is_superclass (clsa, cls, true);
	}

      return retval;
    }
}

inline bool
is_strict_superclass (const cdef_class& clsa, const cdef_class& clsb)
{ return is_superclass (clsa, clsb, false); }

static octave_value_list
class_get_properties (const octave_value_list& args, int nargout)
{
  octave_value_list retval;

  if (args.length () == 1 && args(0).type_name () == "object")
    {
      cdef_class cls (to_cdef (args(0)));

      retval(0) = cls.get_properties ("protected");
    }

  return retval;
}

static octave_value_list
class_get_methods (const octave_value_list& args, int nargout)
{
  octave_value_list retval;

  if (args.length () == 1 && args(0).type_name () == "object")
    {
      cdef_class cls (to_cdef (args(0)));

      retval(0) = cls.get_methods ("protected");
    }

  return retval;
}

static octave_value_list
class_fromName (const octave_value_list& args, int nargout)
{
  octave_value_list retval;

  if (args.length () == 1)
    {
      std::string name = args(0).string_value ();

      if (! error_state)
	retval(0) = to_ov (lookup_class (name));
      else
	error ("fromName: invalid class name, expected a string value");
    }
  else
    error ("fromName: invalid number of parameters");

  return retval;
}

static octave_value_list
class_fevalStatic (const octave_value_list& args, int nargout)
{
  octave_value_list retval;

  if (args.length () > 1 && args(0).type_name () == "object")
    {
      cdef_class cls (to_cdef (args(0)));

      if (! error_state)
	{
	  std::string meth_name = args(1).string_value ();

	  if (! error_state)
	    {
	      // FIXME: can the context be something else?
	      cdef_method meth = cls.find_method (meth_name, "public");

	      if (meth.ok ())
		{
		  if (meth.is_static ())
		    retval = meth.execute (args.splice (0, 2), nargout);
		  else
		    error ("fevalStatic: method `%s' is not static",
			   meth_name.c_str ());
		}
	      else
		error ("fevalStatic: method not found: %s",
		       meth_name.c_str ());
	    }
	  else
	    error ("fevalStatic: invalid method name, expected a string value");
	}
      error ("fevalStatic: invalid object, expected a meta.class object");
    }
  else
    error ("fevalStatic: invalid arguments");

  return retval;
}

static octave_value_list
class_getConstant (const octave_value_list& args, int nargout)
{
  octave_value_list retval;

  if (args.length () == 2 && args(0).type_name () == "object"
      && args(0).class_name () == "meta.class")
    {
      cdef_class cls = to_cdef (args(0));

      if (! error_state)
	{
	  std::string prop_name = args(1).string_value ();

	  if (! error_state)
	    {
	      // FIXME: can the context be something else?
	      cdef_property prop = cls.find_property (prop_name, "public");

	      if (prop.ok ())
		{
		  if (prop.is_constant ())
		    retval(0) = prop.get_value ();
		  else
		    error ("getConstant: property `%s' is not constant",
			   prop_name.c_str ());
		}
	      else
		error ("getConstant: property not found: %s",
		       prop_name.c_str ());
	    }
	  else
	    error ("getConstant: invalid property name, expected a string value");
	}
      else
	error ("getConstant: invalid object, expected a meta.class object");
    }
  else
    error ("getConstant: invalid arguments");

  return retval;
}

#define META_CLASS_CMP(OP, CLSA, CLSB, FUN) \
static octave_value_list \
class_ ## OP (const octave_value_list& args, int nargout) \
{ \
  octave_value_list retval; \
\
  if (args.length () == 2 \
      && args(0).type_name () == "object" && args(1).type_name () == "object" \
      && args(0).class_name () == "meta.class" && args(1).class_name () == "meta.class") \
    { \
      cdef_class clsa = to_cdef (args(0)); \
\
      cdef_class clsb = to_cdef (args(1)); \
\
      if (! error_state) \
	retval(0) = FUN (CLSA, CLSB); \
      else \
	error (#OP ": invalid objects, expected meta.class objects"); \
    } \
  else \
    error (#OP ": invalid arguments"); \
\
  return retval; \
}

META_CLASS_CMP (lt, clsb, clsa, is_strict_superclass)
META_CLASS_CMP (le, clsb, clsa, is_superclass)
META_CLASS_CMP (gt, clsa, clsb, is_strict_superclass)
META_CLASS_CMP (ge, clsa, clsb, is_superclass)
META_CLASS_CMP (eq, clsa, clsb, operator==)
META_CLASS_CMP (ne, clsa, clsb, operator!=)

static cdef_class
make_class (const std::string& name)
{
  cdef_class cls ("meta.class");

  all_classes[name] = cls;
  cls.put ("ConstructOnLoad", false);
  cls.put ("ContainingPackage", Matrix ());
  cls.put ("Description", std::string ());
  cls.put ("DetailedDescription", std::string ());
  cls.put ("Events", Cell ());
  cls.put ("Hidden", false);
  cls.put ("InferiorClasses", Cell ());
  cls.put ("Methods", Cell ());
  cls.put ("Name", name);
  cls.put ("Properties", Cell ());
  cls.put ("Sealed", true);
  cls.put ("SuperClasses", Cell ());

  return cls;
}

static cdef_property
make_property (const cdef_object& cls, const std::string& name,
	       const octave_value& get_method = Matrix (),
	       const std::string& get_access = "public",
	       const octave_value& set_method = Matrix (),
	       const std::string& set_access = "public")
{
  // FIXME: what about default value?

  cdef_property prop ("meta.property");

  prop.put ("Name", name);
  prop.put ("Description", std::string ());
  prop.put ("DetailedDescription", std::string ());
  prop.put ("Abstract", false);
  prop.put ("Constant", false);
  prop.put ("GetAccess", get_access);
  prop.put ("SetAccess", set_access);
  prop.put ("Dependent", false);
  prop.put ("Transient", false);
  prop.put ("Hidden", false);
  prop.put ("GetObservable", false);
  prop.put ("SetObservable", false);
  prop.put ("GetMethod", get_method);
  prop.put ("SetMethod", set_method);
  prop.put ("DefiningClass", to_ov (cls));

  return prop;
}

inline cdef_property
make_attribute (const cdef_object& cls, const std::string& name)
{
  return make_property (cls, name, Matrix (), "public", Matrix (), "private");
}

static cdef_method
make_method (const cdef_object& cls, const std::string& name, const octave_value& fcn,
	     const std::string& m_access = "public", bool is_static = false)
{
  cdef_method meth ("meta.method");

  meth.put ("Abstract", false);
  meth.put ("Access", m_access);
  meth.put ("DefiningClass", to_ov (cls));
  meth.put ("Description", std::string ());
  meth.put ("DetailedDescription", std::string ());
  meth.put ("Hidden", false);
  meth.put ("Name", name);
  meth.put ("Sealed", true);
  meth.put ("Static", is_static);

  meth.set_function (fcn);

  return meth;
}

inline cdef_method
make_method (const cdef_object& cls, const std::string& name, octave_builtin::fcn ff,
	     const std::string& m_access = "public", bool is_static = false)
{
  return make_method (cls, name, octave_value (new octave_builtin (ff)),
		      m_access, is_static);
}

DEFINE_OCTAVE_ALLOCATOR (octave_object);

int octave_object::t_id (-1);

const std::string octave_object::t_name ("object");

void
octave_object::register_type (void)
{
  t_id = octave_value_typeinfo::register_type
    (octave_object::t_name, "<unknown>", octave_value (new octave_object ()));
}

cdef_class
cdef_object_rep::get_class (void) const
{
  cdef_class cls = lookup_class (class_name ());

  return cls;
}

string_vector
cdef_object_rep::map_keys (void) const
{
  cdef_class cls = get_class ();

  if (cls.ok ())
    return cls.get_names ("public");
  
  return string_vector ();
}

octave_value_list
handle_cdef_object::subsref (const std::string& type,
			     const std::list<octave_value_list>& idx,
			     int nargout, int& skip)
{
  skip = 0;

  cdef_class cls = get_class ();

  octave_value_list retval;

  if (! cls.ok ())
    return retval;

  switch (type[0])
    {
    case '.':
	{
	  std::string name = (idx.front ())(0).string_value ();

	  // FIXME: get the right context
	  std::string context = "public";

	  cdef_method meth = cls.find_method (name, context);

	  if (meth.ok ())
	    {
	      int _nargout = (type.length () > 2 ? 1 : nargout);

	      octave_value_list args;

	      skip = 1;

	      if (type.length () > 1 && type[1] == '(')
		{
		  std::list<octave_value_list>::const_iterator it = idx.begin ();

		  args = *++it;

		  skip++;
		}

	      if (meth.is_static ())
		retval = meth.execute (args, _nargout);
	      else
		{
		  refcount++;
		  retval = meth.execute (cdef_object (this), args, _nargout);
		}
	    }

	  if (skip == 0)
	    {
	      cdef_property prop = cls.find_property (name, context);

	      if (prop.ok ())
		{
		  refcount++;
		  retval(0) = prop.get_value (cdef_object (this));

		  skip = 1;
		}
	      else
		error ("subsref: unknown method or property: %s", name.c_str ());
	    }
	  break;
	}
    default:
      error ("object cannot be indexed with `%c'", type[0]);
      break;
    }

  return retval;
}

static void
get_properties_or_methods (const cdef_class& cls, const std::string& nm,
			   const std::string& nm_access,
			   std::list<octave_value>& l)
{
  Cell c = cls.get (nm).cell_value ();

  for (int i = 0; i < c.numel (); i++)
    {
      cdef_object obj = to_cdef (c(i));

      std::string acc = obj.get (nm_access).string_value ();

      if (acc != "private")
	l.push_back (c(i));
    }

  c = cls.get ("SuperClasses").cell_value ();

  for (int i = 0; i < c.numel (); i++)
    {
      cdef_class s_cls  = lookup_class (c(i).string_value ());

      if (s_cls.ok ())
	get_properties_or_methods (s_cls, nm, nm_access, l);
      else
	break;
    }
}

cdef_method
cdef_class::cdef_class_rep::find_method (const std::string& nm,
					 const std::string& context)
{
  method_iterator it = method_map.find (nm);

  if (it == method_map.end ())
    {
      // FIXME: look into class directory
    }
  else
    {
      cdef_method& meth = it->second;

      // FIXME: check if method reload needed

      if (meth.ok () && check_access (context, meth.get_access ()))
	return meth;
    }

  // Look into superclasses

  std::string super_context = superclass_access (context);

  Cell super_classes = get ("SuperClasses").cell_value ();

  for (int i = 0; i < super_classes.numel (); i++)
    {
      cdef_class cls = lookup_class (super_classes(i).string_value ());

      if (! error_state)
	{
	  cdef_method meth = cls.find_method (nm, super_context);

	  if (meth.ok ())
	    return meth;
	}
    }

  return cdef_method ();
}

void
cdef_class::cdef_class_rep::install_method (const cdef_method& meth)
{
  method_map[meth.get_name ()] = meth;
}

void
cdef_class::cdef_class_rep::load_all_methods (void)
{
  // FIXME: re-scan class directory
}

Cell
cdef_class::cdef_class_rep::get_methods (const std::string& context)
{
  load_all_methods ();

  std::list<octave_value> l;

  method_const_iterator it;

  for (it = method_map.begin (); it != method_map.end (); ++it)
    if (check_access (context, it->second.get_access ()))
      l.push_back (to_ov (it->second));

  Cell c (l.size (), 1);

  std::list<octave_value>::const_iterator lit = l.begin ();

  for (int i = 0; lit != l.end(); ++i, ++lit)
    c(i) = (*lit);

  // Look into superclasses

  std::string super_context = superclass_access (context);

  Cell super_classes = get ("SuperClasses").cell_value ();

  for (int i = 0; i < super_classes.numel (); i++)
    {
      cdef_class cls = lookup_class (super_classes(i).string_value ());

      if (! error_state)
	{
	  Cell sc = cls.get_methods (super_context);

	  int n = c.numel ();

	  c.resize_no_fill (n + sc.numel ());
	  c.insert (sc, n, 0);
	}
      else
	break;
    }

  return c;
}

cdef_property
cdef_class::cdef_class_rep::find_property (const std::string& nm,
					   const std::string& context)
{
  property_iterator it = property_map.find (nm);

  if (it != property_map.end ())
    {
      cdef_property& prop = it->second;

      if (prop.ok ()
	  && (check_access (context, prop.get_get_access ()) ||
	      check_access (context, prop.get_set_access ())))
	return prop;
    }

  // Look into superclasses

  std::string super_context = superclass_access (context);

  Cell super_classes = get ("SuperClasses").cell_value ();

  for (int i = 0; i < super_classes.numel (); i++)
    {
      cdef_class cls = lookup_class (super_classes(i).string_value ());

      if (! error_state)
	{
	  cdef_property prop = cls.find_property (nm, super_context);

	  if (prop.ok ())
	    return prop;
	}
    }

  return cdef_property ();
}

void
cdef_class::cdef_class_rep::install_property (const cdef_property& prop)
{
  property_map[prop.get_name ()] = prop;
}

Cell
cdef_class::cdef_class_rep::get_properties (const std::string& context)
{
  std::list<octave_value> l;

  property_const_iterator it;

  for (it = property_map.begin (); it != property_map.end (); ++it)
    if (check_access (context, it->second.get_get_access ()))
      l.push_back (to_ov (it->second));

  Cell c (l.size (), 1);

  std::list<octave_value>::const_iterator lit = l.begin ();

  for (int i = 0; lit != l.end(); ++i, ++lit)
    c(i) = (*lit);

  // Look into superclasses

  std::string super_context = superclass_access (context);

  Cell super_classes = get ("SuperClasses").cell_value ();

  for (int i = 0; i < super_classes.numel (); i++)
    {
      cdef_class cls = lookup_class (super_classes(i).string_value ());

      if (! error_state)
	{
	  Cell sc = cls.get_properties (super_context);

	  int n = c.numel ();

	  c.resize_no_fill (n + sc.numel ());
	  c.insert (sc, n, 0);
	}
      else
	break;
    }

  return c;
}

string_vector
cdef_class::cdef_class_rep::get_names (const std::string& context)
{
  load_all_methods ();

  std::list<std::string> l;

  for (method_const_iterator it = method_map.begin ();
       it != method_map.end(); ++it)
    if (check_access (context, it->second.get_access ()))
      l.push_back (it->second.get_name ());

  for (property_const_iterator it = property_map.begin ();
       it != property_map.end (); ++it)
    if (check_access (context, it->second.get_get_access ()))
      l.push_back (it->second.get_name ());

  string_vector v (l);

  // Look into superclasses

  std::string super_context = superclass_access (context);

  Cell super_classes = get ("SuperClasses").cell_value ();

  for (int i = 0; i < super_classes.numel (); i++)
    {
      cdef_class cls = lookup_class (super_classes(i).string_value ());

      if (! error_state)
	{
	  string_vector sv = cls.get_names (super_context);

	  v.append (sv);
	}
      else
	break;
    }

  return v.qsort (true);
}

octave_value
cdef_property::cdef_property_rep::get_value (const cdef_object& obj)
{
  // FIXME: should check whether we're already in get accessor method

  octave_value retval;
 
  octave_value get_fcn = get ("GetMethod");

  std::string get_access = get ("GetAccess").string_value ();

  if (get_access != "public")
    {
      // FIXME: should check the current call stack
    }

  if (get_fcn.is_empty ())
    retval = obj.get (get ("Name").string_value ());
  else
    {
      octave_value_list args;

      args(0) = to_ov (obj);
      
      args = get_fcn.do_multi_index_op (1, args);

      if (! error_state)
	retval = args(0);
    }

  return retval;
}

void
cdef_method::cdef_method_rep::check_method (void)
{
  // FIXME: check whether re-load is needed
}

octave_value_list
cdef_method::cdef_method_rep::execute (const octave_value_list& args,
				       int nargout)
{
  octave_value_list retval;

  if (! get ("Abstract").bool_value ())
    {
      check_method ();

      if (function.is_defined ())
	{
	  retval = function.do_multi_index_op (nargout, args);
	}
    }
  else
    error ("%s: cannot execute abstract method",
	   get ("Name").string_value ().c_str ());

  return retval;
}

octave_value_list
cdef_method::cdef_method_rep::execute (const cdef_object& obj,
				       const octave_value_list& args,
				       int nargout)
{
  octave_value_list retval;

  if (! get ("Abstract").bool_value ())
    {
      check_method ();

      octave_value_list new_args;

      if (function.is_defined ())
	{
	  new_args.resize (args.length () + 1);

	  new_args(0) = to_ov (obj);
	  for (int i = 0; i < args.length (); i++)
	    new_args(i+1) = args(i);

	  retval = function.do_multi_index_op (nargout, new_args);
	}
    }
  else
    error ("%s: cannot execute abstract method",
	   get ("Name").string_value ().c_str ());

  return retval;
}

octave_value_list
octave_object::subsref (const std::string& type,
			const std::list<octave_value_list>& idx,
			int nargout)
{
  int skip = 0;
  octave_value_list retval;

  // FIXME: should check "subsref" method first

  retval = object.subsref (type, idx, nargout, skip);

  if (! error_state)
    {
      if (type.length () > skip && idx.size () > skip)
	retval = retval(0).next_subsref (nargout, type, idx, skip);
    }

  return retval;
}

void
init_classdef (void)
{
  octave_object::register_type ();

  /* meta classes */
  cdef_class meta_class = make_class ("meta.class");
  cdef_class meta_property = make_class ("meta.property");
  cdef_class meta_method = make_class ("meta.method");
  cdef_class meta_event = make_class ("meta.event");
  cdef_class meta_package = make_class ("meta.package");
  cdef_class meta_dynproperty = make_class ("meta.dynamicproperty");

  /* meta.class properties */
  meta_class.install_property (make_attribute (meta_class, "ConstructOnLoad"));
  meta_class.install_property (make_property  (meta_class, "ContainingPackage"));
  meta_class.install_property (make_property  (meta_class, "Description"));
  meta_class.install_property (make_property  (meta_class, "DetailedDescription"));
  meta_class.install_property (make_property  (meta_class, "Events"));
  meta_class.install_property (make_attribute (meta_class, "Hidden"));
  meta_class.install_property (make_attribute (meta_class, "InferiorClasses"));
  meta_class.install_property
      (make_property  (meta_class, "Methods",
		       octave_value (new octave_builtin (class_get_methods)),
		       "public", Matrix (), "private"));
  meta_class.install_property (make_attribute (meta_class, "Name"));
  meta_class.install_property
      (make_property  (meta_class, "Properties",
		       octave_value (new octave_builtin (class_get_properties)),
		       "public", Matrix (), "private"));
  meta_class.install_property (make_attribute (meta_class, "Sealed"));
  meta_class.install_property (make_attribute (meta_class, "SuperClasses"));
  /* meta.class methods */
  meta_class.install_method (make_method (meta_class, "fromName", class_fromName,
					  "public", true));
  meta_class.install_method (make_method (meta_class, "fevalStatic", class_fevalStatic,
					  "public", false));
  meta_class.install_method (make_method (meta_class, "getConstant", class_getConstant,
					  "public", false));
  meta_class.install_method (make_method (meta_class, "eq", class_eq));
  meta_class.install_method (make_method (meta_class, "ne", class_ne));
  meta_class.install_method (make_method (meta_class, "lt", class_lt));
  meta_class.install_method (make_method (meta_class, "le", class_le));
  meta_class.install_method (make_method (meta_class, "gt", class_gt));
  meta_class.install_method (make_method (meta_class, "ge", class_ge));

  /* meta.method properties */
  meta_method.install_property (make_attribute (meta_method, "Abstract"));
  meta_method.install_property (make_attribute (meta_method, "Access"));
  meta_method.install_property (make_attribute (meta_method, "DefiningClass"));
  meta_method.install_property (make_attribute (meta_method, "Description"));
  meta_method.install_property (make_attribute (meta_method, "DetailedDescription"));
  meta_method.install_property (make_attribute (meta_method, "Hidden"));
  meta_method.install_property (make_attribute (meta_method, "Name"));
  meta_method.install_property (make_attribute (meta_method, "Sealed"));
  meta_method.install_property (make_attribute (meta_method, "Static"));

  /* met.property properties */
  meta_property.install_property (make_attribute (meta_property, "Name"));
  meta_property.install_property (make_attribute (meta_property, "Description"));
  meta_property.install_property (make_attribute (meta_property, "DetailedDescription"));
  meta_property.install_property (make_attribute (meta_property, "Abstract"));
  meta_property.install_property (make_attribute (meta_property, "Constant"));
  meta_property.install_property (make_attribute (meta_property, "GetAccess"));
  meta_property.install_property (make_attribute (meta_property, "SetAccess"));
  meta_property.install_property (make_attribute (meta_property, "Dependent"));
  meta_property.install_property (make_attribute (meta_property, "Transient"));
  meta_property.install_property (make_attribute (meta_property, "Hidden"));
  meta_property.install_property (make_attribute (meta_property, "GetObservable"));
  meta_property.install_property (make_attribute (meta_property, "SetObservable"));
  meta_property.install_property (make_attribute (meta_property, "GetMethod"));
  meta_property.install_property (make_attribute (meta_property, "SetMethod"));
  meta_property.install_property (make_attribute (meta_property, "DefiningClass"));
  /* meta.property events */
  // FIXME: add events
}

DEFUN_DLD (__init_classdef__, , , "")
{
  init_classdef ();

  return octave_value ();
}

DEFUN_DLD (get_class, args, , "")
{
  octave_value retval;

  if (args.length () == 1)
    {
      std::string cname = args(0).string_value ();

      if (! error_state)
	retval = to_ov (lookup_class (cname));
      else
	error ("invalid class name, expected a string value");
    }
  else
    print_usage ();

  return retval;
}

/*
;;; Local Variables: ***
;;; mode: C++ ***
;;; End: ***
*/
